function ToHexString(v, l){
  l = (typeof(l) === 'number' && (l === 8 || l === 16)) ? l : 8;
  var r = v.toString(16);
  if (r.length < 2)
    r = "0" + r;
  if (l === 16 && r.length < 4)
    r = ((r.length === 2) ? "00" : "0") + r;
  return r;
}

function GetStrings(s, rc){
  s = (typeof(s) === 'string') ? [s] : s;
  var m = s[0].match(/"((?:\\.|[^"\\])*)"/);
  if (m && m.length > 0){
    let i = s[0].indexOf(m[0]);
    if (i > 0){
      s[0] = s[0].substr(0, i) + rc + s[0].substr(i + m[0].length);
      s.push(m[0]);
      return GetStrings(s, rc);
    }
  }
  return s;
}

function Tokenize(l){
  var m = l.match(/([^()]+)|([()])/g);
  var ip = false; 
  var t = [];
  m.forEach((mi)=>{
    if (mi === "("){
      ip = true;
    } else if (mi === ")"){
      ip = false;
    } else {
      if (ip){
        mi.split(/[\s,]+/).forEach((i)=>{
          if (i !== "")
            t.push("+" + i);
        });
      } else {
        var strs = GetStrings(mi, "*");
        var stri = 1;
        strs[0].split(/[\s,]+/).forEach((i)=>{
          if (i === "*" && stri < strs.length){
            t.push(strs[stri].substr(1, strs[stri].length - 2));
            stri += 1;
          } else if (i !== ""){
            t.push(i);
          }
        });
      }
    }
  });
  return t;
}



function StoreVarLabel(data, lbl, val){
  if (lbl in data.varlabels)
    throw new Error("Variable label '" + lbl + "' defined more than once.");
  if (lbl in data.jmplabels)
    throw new Error("Variable label '" + lbl + "' matches already existing jump label.");
  let v = Number.NaN;
  if (val[0] === "$"){
    if (val.length >= 2 && val.length <= 5)
      v = parseInt(val.substr(1), 16);
  } else if (val[0] === "#"){
    if (val.startsWith("#b")){
      v = parseInt(val.substr(2), 2);
    } else {
      v = parseInt(val);
    }
  }
  if (!isNaN(v)){
    if (v < 0 || v > 65535)
      throw new Error("Variable '" + lbl + "' value is out of bounds for this processor.");
    data.varlabels[lbl] = v;
  } else {
    throw new Error("Variable '" + lbl + "' value malformed.");
  }
}

function StoreJmpLabel(data, lbl){ 
  if (lbl in data.jmplabels)
    throw new Error("Jump label '" + lbl + "' defined more than once at program address " + ToHexString(data.PC));
  if (lbl in data.varlabels)
    throw new Error("Jump label '" + lbl + "' matches already existing variable name at program address " + ToHexString(data.PC)); 
  data.jmplabels[lbl] = data.PC;
}



function AddrModeVal(data, tA, tB){
  var mode = "";
  var v = Number.NaN;
  if (tA[0] === '#'){
    mode = "i"; // Immediate
    if (tA[1] === '$'){
      v = parseInt(tA.substr(2), 16);
    } else if (tA[1] === 'b'){
      v = parseInt(tA.substr(2), 2);
    } else {
      v = parseInt(tA.substr(1));
      if (isNaN(v)){
        let lbl = tA.substr(1);
        if (lbl in this.__varlabels){
          if (this.__varlabels[lbl] < 256)
            v = this.__varlabels[lbl]
        }
      }
    }
  } else if (tA[0] === '$'){
    v = parseInt(tA.substr(1), 16);
    if (tA.length === 3){
      // Zero Page
      if (tB !== null){
        if (tB === "X")
          mode = "zX";
        if (tB === "Y")
          mode = "zY";
      } else {
        mode = "z";
      }
    } else if (tA.length === 5){
      // Absolute
      if (tB !== null){
        if (tB === "X"){
          mode = "aX";
        } else if (tB === "Y"){
          mode = "aY";
        }
      } else {
        mode = "a";
      }
    }
  } else if (tA.startsWith("+")){
    if (tB !== null){
      if (tB === "+X"){ // The plus exists due to how tokenize() works.
        mode = "nX";
      } else if (tB === "Y"){
        mode = "nY";
      }

      if (mode !== ""){
        if (tA.startsWith("+$")){
          v = parseInt(tA.substr(2), 16);
        } else {
          let lbl = tA.substr(1);
          if (lbl in this.__varlabels){
            if (this.__varlabels[lbl] < 256)
              v = this.__varlabels[lbl];
          }
        }
      }
    }
  } else {
    // We've been given a variable name. Need to look it up.
    // if the var is 1 byte, assume "Zero Page". If var 2 bytes, assume "Absolute"
    if (tA in data.varlabel){
      v = data.varlabel[tA];
      if (v < 256){
        if (tB !== null){
          if (tB === "X")
            mode = "zX";
        } else {
          mode = "z";
        }
      } else {
        if (tB !== null){
          if (tB === "X"){
            mode = "aX";
          } else if (tB === "Y"){
            mode = "aY";
          }
        } else {
          mode = "a";
        }
      }
    }
  }

  if (mode !== "" && ! isNaN(v))
    return [mode, v];
  throw new Error("Malformed op-code or value on program address " + ToHexString(data.PC));
}


// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------

function StoreSingleOp(data, cmp, code){
  if (cmp.tokens.length === 1){
    cmp.op.push(code);
    data.PC += 1;
    return false;
  }
  return true;
};

function StoreBranchOp(data, cmp, code){
  if (cmp.tokens.length === 2){
    let v = Number.NaN;
    if (cmp.tokens[1][0] === '$'){
      if (cmp.tokens[1].length === 3 || cmp.tokens[1].length === 5){
        v = parseInt(cmp.tokens[1].substr(1), 16);
      }
    } else {
      if (cmp.tokens[1] in data.jmplabels){
        v = data.jmplabels[cmp.tokens[1]];
      }
    }

    if (!isNaN(v)){
      if (v < (data.PC+1) - 127 || v > (data.PC+1) + 128)
        throw new Error("Branch exceeds maximum number of bytes on program address " + ToHexString(data.PC + 1));
      v = v - (data.PC + 1);
      cmp.op.push(code);
      cmp.op.push((v >= 0) ? v : v + 255);
      data.PC += 2;
      return false;
    }
  }
  return true;
};

function StoreOp(data, cmp, codes, mint, maxt, zpm){
  let tokens = cmp.tokens;
  zpm = (zpm === "zY") ? "zY" : "zX";
  if (cmp.tokens.length >= mint && cmp.tokens.length <= maxt){
    let mv = AddrModeVal(data, cmp.tokens[1], (cmp.tokens.length === 3) ? cmp.tokens[2].toUpperCase() : null);
    let modena = true;
    switch(mv[0]){
      case "i":
        modena = codes[0] === null;
        if (!modena){
          cmp.op.push(codes[0]);
          cmp.op.push(mv[1]);
          data.PC += 2;
        } break;
      case "z":
        modena = codes[1] === null;
        if (!modena){
          cmp.op.push(codes[1]);
          cmp.op.push(mv[1]);
          data.PC += 2;
        } break;
      case "zY":
      case "zX":
        modena = (mv[0] === zpm && codes[2] === null);
        if (!modena){
          cmp.op.push(codes[2]);
          cmp.op.push(mv[1]);
          data.PC += 2;
        } break;
      case "a":
        modena = codes[3] === null;
        if (!modena){
          cmp.op.push(codes[3]);
          cmp.op.push(mv[1] & 0x000000FF);
          cmp.op.push((mv[1] & 0x0000FF00) >> 8);
          data.PC += 3;
        } break;
      case "aX":
        modena = codes[4] === null;
        if (!modena){
          cmp.op.push(codes[4]);
          cmp.op.push(mv[1] & 0x000000FF);
          cmp.op.push((mv[1] & 0x0000FF00) >> 8);
          data.PC += 3;
        } break;
      case "aY":
        modena = codes[5] === null;
        if (!modena){
          cmp.op.push(codes[5]);
          cmp.op.push(mv[1] & 0x000000FF);
          cmp.op.push((mv[1] & 0x0000FF00) >> 8);
          data.PC += 3;
        } break;
      case "nX":
        modena = codes[6] === null;
        if (!modena){
          cmp.op.push(codes[6]);
          cmp.op.push(mv[1]);
          data.PC += 2;
        } break;
      case "nY":
        modena = codes[7] === null;
        if (!modena){
          cmp.op.push(codes[7]);
          cmp.op.push(mv[1]);
          data.PC += 2;
        } break;
    }
    if (modena)
      throw new Error("Op-code does not support implied mode on program address " + ToHexString(data.PC));
    return false;
  }
  return true;
};


// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------


function directive_DEFINE(data, cmp){
  // Variable label!!
  StoreVarLabel(data, cmp.tokens[1], cmp.tokens[2]);
}


function directive_BYTES(data, cmp){
  // Compiler directive. Treat all proceeding tokens as values and store them raw.
  for(let i=1; i < cmp.tokens.length; i++){
    let v = Number.NaN;
    if (cmp.tokens[i].startsWith("$")){
      v = parseInt(cmp.tokens[i].substr(1), 16);
    } else if (cmp.tokens[i].startsWith("b")){
      v = parseInt(cmp.tokens[i].substr(1), 2);
    } else {
      v = parseInt(cmp.tokens[i]);
    }
    if (isNaN(v) || v < 0 || v >= 256)
      throw new Error("Byte list value is malformed or out of bounds at program address " + ToHexString(data.PC));
    cmp.op.push(v);
    data.PC += 1;
  };
}



// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------


class Assembler{
  constructor(initpc){
    // Inital value for program counter. Used by the reset() method.
    this.__initPC = (typeof(initpc) === 'number' && initpc >= 0) ? initpc : 0;
    this.__data = {
      // Labels that hold variable values.
      varlabels: {},
      // Labels that hold jump/branch locations.
      jmplabels: {}, 
      // Program counter to track where in the code/memory a
      // operation is located.
      PC: this.__initPC,
      // Currently compiled code.
      result: []
    };
    this.__directives = {
      ".bytes":directive_BYTES,
      ".byt":directive_BYTES,
      ".define":directive_DEFINE,
      ".def":directive_DEFINE
    };
  }

  get PC(){return this.__data.PC;}
  get variables(){
    var v = {};
    Object.keys(this.__data.varlabels).forEach((lbl)=>{
      if (this.__data.varlabels.hasOwnProperty(lbl))
        v[lbl] = this.__data.varlabels[lbl];
    });
    return v;
  }
  get jumplabels(){
    var j = {};
    Object.keys(this.__data.jmplabels).forEach((lbl)=>{
      if (this.__data.jmplabels.hasOwnProperty(lbl))
        j[lbl] = this.__data.jmplabels[lbl];
    });
    return j;
  }


  compile(src){
    var cmp = {
      op: [],
      tokens: []
    };
    src.split("\n").forEach((line)=>{
      line = line.trim();
      if (line === ""){return this;}
      if (line[0] !== ";"){ // Skip comment lines.
        line = line.split(";")[0].trim(); // Take out any trailing comments.
        // Checking for any jump labels!
        if (line.indexOf(":") >= 0){
          let s = line.split(":");
          let lbl = s[0].trim();
          if (lbl.length <= 0)
            throw new Error("Malformatted jump label at program address " + ToHexString(this.__data.PC));
          StoreJmpLabel(this.__data, lbl);
          line = s[1].trim();
          if (line.length <= 0){return this;} // Nothing left to process.
        }
        // Finally... tokenize the main command.
        cmp.tokens = Tokenize(line);

        var directivesHandled = false;
        if (cmp.tokens[0].startsWith(".")){
          let d = cmp.tokens[0].toLowerCase();
          if (d in this.__directives){
            this.__directives[d](this.__data, cmp);
            directivesHandled = true;
          }
        }

        if (!directivesHandled && cmp.tokens[0].length === 3){ 
          let procFailed = false;
          let mv = null;
          // Possible op code.
          switch(cmp.tokens[0].toLowerCase()){
// --- ADC
            case 'adc':
              procFailed = StoreOp(this.__data, cmp, [0x69, 0x65, 0x75, 0x6D, 0x7D, 0x79, 0x61, 0x71], 2, 3); break;
// --- AND
            case 'and':
              procFailed = StoreOp(this.__data, cmp, [0x29, 0x25, 0x35, 0x2D, 0x3D, 0x39, 0x21, 0x31], 2, 3); 
              break;
// --- ASL
            case 'asl':
              if (cmp.tokens.length === 2){
                if (cmp.tokens[1].toUpperCase() === 'A'){
                  cmp.op.push(0x0A);
                  this.__data.PC += 1;
                  break;
                }
              }
              procFailed = StoreOp(this.__data, cmp, [null, 0x06, 0x16, 0x0E, 0x1E, null, null, null], 2, 3);  
              break;
// --- BCC
            case 'bcc':
              procFailed = StoreBranchOp(this.__data, cmp, 0x90); break;
// --- BCS
            case 'bcs':
              procFailed = StoreBranchOp(this.__data, cmp, 0xB0); break;
// --- BEQ
            case 'beq':
              procFailed = StoreBranchOp(this.__data, cmp, 0xF0); break;
// --- BIT
            case 'bit':
              procFailed = StoreOp(this.__data, cmp, [null, 0x24, null, 0x2C, null, null, null, null], 2, 2); break;
// --- BMI
            case 'bmi':
              procFailed = StoreBranchOp(this.__data, cmp, 0x30); break;
// --- BNE
            case 'bne':
              procFailed = StoreBranchOp(this.__data, cmp, 0xD0); break;
// --- BPL
            case 'bpl':
              procFailed = StoreBranchOp(this.__data, cmp, 0x10); break;
// --- BRK
            case 'brk':
              procFailed = StoreSingleOp(this.__data, cmp, 0x00); break;
// --- BVC
            case 'bvc':
              procFailed = StoreBranchOp(this.__data, cmp, 0x50); break;
// --- BVS
            case 'bvs':
              procFailed = StoreBranchOp(this.__data, cmp, 0x70); break;
// --- CLC
            case 'clc':
              procFailed = StoreSingleOp(this.__data, cmp, 0x18); break;
// --- CLD
            case 'cld':
              procFailed = StoreSingleOp(this.__data, cmp, 0xD8); break;
// --- CLI
            case 'cli':
              procFailed = StoreSingleOp(this.__data, cmp, 0x58); break;
// --- CLV
            case 'clv':
              procFailed = StoreSingleOp(this.__data, cmp, 0xB8); break;
// --- CMP
            case 'cmp':
              procFailed = StoreOp(this.__data, cmp, [0xC9, 0xC5, 0xD5, 0xCD, 0xDD, 0xD9, 0xC1, 0xD1], 2, 3); break;
// --- CPX
            case 'cpx':
              procFailed = StoreOp(this.__data, cmp, [0xE0, 0xE4, null, 0xEC, null, null, null, null], 2, 2); break;
// --- CPY
            case 'cpy':
              procFailed = StoreOp(this.__data, cmp, [0xC0, 0xC4, null, 0xCC, null, null, null, null], 2, 2); break; 
// --- DEC
            case 'dec':
              procFailed = StoreOp(this.__data, cmp, [null, 0xC6, 0xD6, 0xCE, 0xDE, null, null, null], 2, 3); break; 
// --- DEX
            case 'dex':
              procFailed = StoreSingleOp(this.__data, cmp, 0xCA); break;
// --- DEY
            case 'dey':
              procFailed = StoreSingleOp(this.__data, cmp, 0x88); break;
// --- EOR
            case 'eor':
              procFailed = StoreOp(this.__data, cmp, [0x49, 0x45, 0x55, 0x4D, 0x5D, 0x59, 0x41, 0x51], 2, 3); break; 
// --- INC
            case 'inc':
              procFailed = StoreOp(this.__data, cmp, [null, 0xE6, 0xF6, 0xEE, 0xFE, null, null, null], 2, 3); break; 
// --- INX
            case 'inx':
              procFailed = StoreSingleOp(this.__data, cmp, 0xE8); break;
// --- INY
            case 'iny':
              procFailed = StoreSingleOp(this.__data, cmp, 0xC8); break;
// --- JMP
            case 'jmp':
              if (cmp.tokens.length === 2){
                let v = Number.NaN;
                let code = (cmp.tokens[1].startsWith("+")) ? 0x6C : 0x4C;
                if (cmp.tokens[1].startsWith("$")){
                  if (cmp.tokens[1].length === 5)
                    v = parseInt(cmp.tokens[1].substr(1), 16);
                } else if (cmp.tokens[1].startsWith("+$")){
                  if (cmp.tokens[1].length === 6)
                    v = parseInt(cmp.tokens[1].substr(2), 16);
                } else {
                  let lbl = (cmp.tokens[1].startsWith("+")) ? cmp.tokens[1].substr(1) : cmp.tokens[1];
                  if (lbl in this.__data.jmplabels)
                    v = this.__data.jmplabels[lbl];
                }
                if (!isNaN(v)){
                  cmp.op.push(code);
                  cmp.op.push(v & 0x000000FF);
                  cmp.op.push((v & 0x0000FF00) >> 8);
                  this.__data.PC += 3;
                } else {
                  throw new Error("Malformed op-code or value on program address " + ToHexString(this.__data.PC));
                }
              } else { procFailed = true; }
              break;
// --- JSR
            case 'jsr':
              if (cmp.tokens.length === 2){
                let v = Number.NaN;
                if (cmp.tokens[1].startsWith("$")){
                  if (cmp.tokens[1].length === 5)
                    v = parseInt(cmp.tokens[1].substr(1), 16);
                } else {
                  if (cmp.tokens[1] in this.__data.jmplabels)
                    v = this.__data.jmplabels[cmp.tokens[1]];
                }
                if (!isNaN(v)){
                  cmp.op.push(0x20);
                  cmp.op.push(v & 0x000000FF);
                  cmp.op.push((v & 0x0000FF00) >> 8);
                  this.__data.PC += 3;
                } else {
                  throw new Error("Malformed op-code or value on program address " + ToHexString(this.__data.PC));
                }
              } else {
                procFailed = true;
              }
              break;
// --- LDA
            case 'lda':
              procFailed = StoreOp(this.__data, cmp, [0xA9, 0xA5, 0xB5, 0xAD, 0xBD, 0xB9, 0xA1, 0xB1], 2, 3); break;
// --- LDX
            case 'ldx':
              procFailed = StoreOp(this.__data, cmp, [0xA2, 0xA6, 0xB6, 0xAE, null, 0xBE, null, null], 2, 3, "zY"); break;
// --- LDY
            case 'ldy':
              procFailed = StoreOp(this.__data, cmp, [0xA0, 0xA4, 0xB4, 0xAC, 0xBC, null, null, null], 2, 3); break;
// --- LSR
            case 'lsr':
              if (cmp.tokens.length === 2){
                if (cmp.tokens[1].toUpperCase() === 'A'){
                  cmp.op.push(0x4A);
                  this.__data.PC += 1;
                  break;
                }
              }
              procFailed = StoreOp(this.__data, cmp, [null, 0x46, 0x56, 0x4E, 0x5E, null, null, null], 2, 3); break;
// --- NOP
            case 'nop':
              procFailed = StoreSingleOp(this.__data, cmp, 0xEA); break;
// --- ORA
            case 'ora':
              procFailed = StoreOp(this.__data, cmp, [0x09, 0x05, 0x15, 0x0D, 0x1D, 0x19, 0x01, 0x11], 2, 3); break;
// --- PHA
            case 'pha':
              procFailed = StoreSingleOp(this.__data, cmp, 0x48); break;
// --- PHP
            case 'php':
              procFailed = StoreSingleOp(this.__data, cmp, 0x08); break;
// --- PLA
            case 'pla':
              procFailed = StoreSingleOp(this.__data, cmp, 0x68); break;
// --- PLP
            case 'plp':
              procFailed = StoreSingleOp(this.__data, cmp, 0x28); break;
// --- ROL
            case 'rol':
              if (cmp.tokens.length === 2){
                if (cmp.tokens[1].toUpperCase() === 'A'){
                  cmp.op.push(0x2A);
                  this.__data.PC += 1;
                  break;
                }
              }
              procFailed = StoreOp(this.__data, cmp, [null, 0x26, 0x36, 0x2E, 0x3E, null, null, null], 2, 3); break;
// --- ROR
            case 'ror':
              if (cmp.tokens.length === 2){
                if (cmp.tokens[1].toUpperCase() === 'A'){
                  cmp.op.push(0x6A);
                  this.__data.PC += 1;
                  break;
                }
              }
              procFailed = StoreOp(this.__data, cmp, [null, 0x66, 0x76, 0x6E, 0x7E, null, null, null], 2, 3); break;
// --- RTI
            case 'rti':
              procFailed = StoreSingleOp(this.__data, cmp, 0x40); break;
// --- RTS
            case 'rts':
              procFailed = StoreSingleOp(this.__data, cmp, 0x60); break;
              break;
// --- SBC
            case 'sbc':
              procFailed = StoreOp(this.__data, cmp, [0xE9, 0xE5, 0xF5, 0xED, 0xFD, 0xF9, 0xE1, 0xF1], 2, 3); break;
// --- SEC
            case 'sec':
              procFailed = StoreSingleOp(this.__data, cmp, 0x38); break;
// --- SED
            case 'sed':
              procFailed = StoreSingleOp(this.__data, cmp, 0xF8); break;
// --- SEI
            case 'sei':
              procFailed = StoreSingleOp(this.__data, cmp, 0x78); break;
              break;
// --- STA
            case 'sta':
              procFailed = StoreOp(this.__data, cmp, [null, 0x85, 0x95, 0x8D, 0x9D, 0x99, 0x81, 0x91], 2, 3); break;
// --- STX
            case 'stx':
              procFailed = StoreOp(this.__data, cmp, [null, 0x86, 0x96, 0x8E, null, null, null, null], 2, 3, "zY"); break;
// --- STY
            case 'sty':
              procFailed = StoreOp(this.__data, cmp, [null, 0x84, 0x94, 0x8C, null, null, null, null], 2, 3); break;
// --- TAX
            case 'tax':
              procFailed = StoreSingleOp(this.__data, cmp, 0xAA); break;
// --- TAY
            case 'tay':
              procFailed = StoreSingleOp(this.__data, cmp, 0xA8); break;
// --- TSX
            case 'tsx':
              procFailed = StoreSingleOp(this.__data, cmp, 0xBA); break;
// --- TXA
            case 'txa':
              procFailed = StoreSingleOp(this.__data, cmp, 0x8A); break;
// --- TXS
            case 'txs':
              procFailed = StoreSingleOp(this.__data, cmp, 0x9A); break;
// --- TYA
            case 'tya':
              procFailed = StoreSingleOp(this.__data, cmp, 0x98); break;
// --- ---
// EOL
// --- ---
            default:
              throw new Error("Unknown op-code '" + cmp.tokens[0].toUpperCase() + "' at program address " + ToHexString(this.__data.PC));
          }

          if (procFailed)
            throw new Error("Malformed op-code on program address " + ToHexString(this.__data.PC));
        } else {
          if (!directivesHandled)
            throw new Error("Failed to compile line '" + line + "' at program address " + ToHexString(this.__data.PC));
        }
      }
    });
    this.__data.result = this.__data.result.concat(cmp.op);
    return this;
    //return new Uint8Array(op);
  }

  result(){
    return new Uint8Array(this.__data.result);
  }

  reset(){
    this.__data.varlabels = {};
    this.__data.jmplabels = {}; 
    this.__data.PC = this.__initPC;
    this.__data.result = [];
    return this;
  }
}


module.exports = Assembler;
