
class Clock{
  constructor(){
    this.__ticktime = 1000;
    this.__dt = 0;
    this.__ltime = 0;
    this.__out = [];
  }

  get hz(){return Math.floor(1000 / this.__ticktime);}
  set hz(hz){
    if(hz >= 1){
      this.__ticktime = 1000 / hz;
      this.__dt = 0;
      this.__ltime = 0;
    }
  }

  clk(fn){
    if (typeof(fn) === 'function')
      this.__out.push(fn);
    return this;
  }

  tick(){
    let ctime = (new Date()).getTime();
    if(this.__ltime !== 0){
      this.__dt += ctime - this.__ltime;
      while(this.__dt > this.__ticktime){
        this.__out.forEach((fn)=>{fn();});
        this.__dt -= this.__ticktime
      }
    } else {
      // Call all of the callbacks immediately. This is the first
      // legit tick!
      this.__out.forEach((fn)=>{fn();});
    }
    this.__ltime = ctime;
    return this;
  }
}


module.exports = Clock;
