

class IMem{
  constructor(){}

  get pages(){return 0;}
  get size(){return 0;}
  get writable(){return false;}

  get address(){return 0;}
  set address(a){}

  get byte(){return -1;}
  set byte(b){}

  // Shorthand method for accessing an address and returning a byte.
  // Same as...
  //
  // mem.address = <16 bit value>
  // let res = mem.byte
  //
  // NOTE: This will trigger read listeners.
  read(a){
    this.address = a;
    return this.byte;
  }

  // Shorthand method for accessing an address and setting it's byte.
  // Same as...
  //
  // mem.address = <16 bit value>
  // mem.byte = <1 byte value>
  //
  // NOTE: This will trigger write listeners.
  write(a, b){
    if (this.writable){
      this.address = a;
      this.byte = b & 0xFF;
    }
    return this;
  }

  // Same as .read() but will NOT trigger read listeners.
  peek(a){return 0;}
  // Same as .write() but will NOT trigger write listeners.
  poke(a, b){return this;}

  // Starting at <address> will fill memory (regardless of whether
  // it's ROM or RAM) with the values in <data>
  // NOTE: <data> is assumed to only contain byte values.
  load(address, data){
    return this;
  }

  // Clears 256 bytes of memory starting from <page> * 256
  // NOTE: This affects both RAM and ROM.
  clearPage(page){
    return this;
  }

  // Clears all memory
  // NOTE: This affects both RAM and ROM.
  clear(){
    return this;
  }
}


module.exports = IMem;
